unit cxdfuncs;

interface
uses
  Classes, cxdconsts
  ;

type
  tIntFunc = function: integer;
  tStrFunc = function: string;

  TFunctionsRec = packed record
    fnPos, fnCount: tIntFunc;
    fnReplace: tStrFunc;
    fnPosFile, fnCountFile, fnReplaceFile: tIntFunc;
  end;

  TContestantFunctions = packed array[TContestant] of TFunctionsRec;

var { Global }
  Self: TComponent;
  _S_, _R_, _SubStr_, _Replacement_, _FileName_, _RFilename_: string;
  //_FileNames_: array of string;
  _CaseSensitive_, _IgnoreCase_: boolean;
  _StartPos_, _LoopCount_: integer;
  ContestantFunctions: TContestantFunctions;
  _SLength, _SubStrLength: integer;

procedure initContestantfunctions;
implementation
uses sysutils, FastStrings, FastStringfuncs, mwtSearch, Searches, cxpos;


function DelphiPos: integer; forward; overload;
function DelphiPos(const Text: string): integer; forward; overload;
function DelphiPos(const StartPos: integer): integer; forward; overload;
function DelphiPos(const Text: string; const StartPos: integer): integer; forward; overload;
function DelphiCount: integer; forward; overload;
function DelphiCount(const Text: string): integer; forward; overload;
function DelphiReplace: string; forward; overload;
function DelphiReplace(const Text: string): string; forward; overload;
function DelphiPosFile: integer; forward;
function DelphiCountFile: integer; forward;
function DelphiReplaceFile: integer; forward;
function FastStringPos: integer; forward;

function FastStringCount: integer; overload; forward;

function FastStringReplace: string; overload; forward;
function FastStringPosFile: integer; forward;
function FastStringCountFile: integer; forward;
function FastStringReplaceFile: integer; forward;
function AJSearchPos: integer; forward;
function AJSearchCount: integer; forward;
function AJSearchReplace: string; forward; overload;
function AJSearchReplace(const _S_: string): string; forward; overload;
function AJSearchPosFile: integer; forward;
function AJSearchCountFile: integer; forward;
function AJSearchReplaceFile: integer; forward;
function MWSearchPos: integer; forward; overload;
//function MWSearchPos(const S: string): integer; forward; overload;
//function MWSearchPos(const mw: TmTSearch; const S: string): integer; forward; overload;
function MWSearchCount: integer; forward; overload;
function MWSearchReplace: string; forward; overload;
function MWSearchPosFile: integer; forward;
function MWSearchCountFile: integer; forward;
function MWSearchReplaceFile: integer; forward;
function cxposPos: integer; forward;
function cxposCount: integer; forward;
function cxposReplace: string; forward;
function cxposPosFile: integer; forward;
function cxposCountFile: integer; forward;
function cxposReplaceFile: integer; forward;

//function DelphiPosIgnoreCase: integer; //(const SubStr, S: string): integer;
//begin
//  Result := Pos(uppercase(_SubStr_), uppercase(_S))
//end;

// IMPLEMENTATIONS

const
  //default read/write file access
  default_fmRead = fmOpenRead or fmShareDenyNone;
  default_fmWrite = fmOpenReadWrite or fmShareDenyNone;
  // fmShareDenyWrite actually better choice,
  // but too often conflict with other app

function DelphiPos: integer; overload;
begin
  if _StartPos_ <= 1 then
    if _CaseSensitive_ then
      Result := Pos(_SubStr_, _S_)
    else
      Result := Pos(uppercase(_SubStr_), uppercase(_S_))
  else begin
    if _CaseSensitive_ then
      Result := Pos(_SubStr_, copy(_S_, _StartPos_, MAXINT))
    else
      Result := Pos(uppercase(_SubStr_), uppercase(copy(_S_, _StartPos_, MAXINT)));
    if Result > 0 then inc(Result, _StartPos_ - 1)
  end;
end;

function DelphiPos(const Text: string): integer; overload;
// for StartPos <= 1, file access
begin
  if _CaseSensitive_ then
    Result := Pos(_SubStr_, copy(Text, _StartPos_, MAXINT))
  else
    Result := Pos(uppercase(_SubStr_), uppercase(copy(Text, _StartPos_, MAXINT)));
  if Result > 0 then inc(Result, _StartPos_ - 1) //
end;

function DelphiPos(const Text: string; const StartPos: integer): integer; overload;
// for StartPos > 1, file access
begin
  if _CaseSensitive_ then
    Result := Pos(_SubStr_, copy(Text, StartPos, MAXINT))
  else
    Result := Pos(uppercase(_SubStr_), uppercase(copy(Text, StartPos, MAXINT)));
  if Result > 0 then inc(Result, StartPos - 1) //
end;

function DelphiPos(const StartPos: integer): integer; overload;
// for StartPos > 1
begin
//  Result := DelphiPos(_S_, StartPos);
  if _CaseSensitive_ then
    Result := Pos(_SubStr_, copy(_S_, StartPos, MAXINT))
  else
    Result := Pos(uppercase(_SubStr_), uppercase(copy(_S_, StartPos, MAXINT)));
  if Result > 0 then inc(Result, StartPos - 1) //
end;

function DelphiCount: integer;
var
  fPos: integer;
  SubLen: integer;
begin
  Result := 0;
  fPos := DelphiPos;
  if fPos > 0 then begin
    SubLen := length(_SubStr_);
    // no cheats, we'll do as efficient as possible apart from
    // the real speed of the string matching pattern algorithm
    // hence we have to unroll the computation of substr length
    // out of the loop
    while fPos > 0 do begin
      inc(Result);
      fPos := DelphiPos(fPos + SubLen);
    end;
  end
end;

function DelphiCount(const Text: string): integer;
var
  fPos: integer;
  SubLen: integer;
begin
  Result := 0;
  fPos := DelphiPos(Text);
  if fPos > 0 then begin
    SubLen := length(_SubStr_);
    while fPos > 0 do begin
      inc(Result);
      fPos := DelphiPos(Text, fPos + SubLen);
    end;
  end
end;

function DelphiReplace: string;
var
  rf: TReplaceFlags;
begin
  rf := [rfReplaceAll];
  if _IgnoreCase_ then include(rf, rfIgnoreCase);
  if _StartPos_ = 1 then
    Result := StringReplace(_S_, _SubStr_, _Replacement_, rf)
  else
    Result := StringReplace(copy(_S_, _StartPos_, MAXINT), _SubStr_, _Replacement_, rf);
end;

function DelphiReplace(const Text: string): string;
var
  rf: TReplaceFlags;
begin
  rf := [rfReplaceAll];
  if _IgnoreCase_ then include(rf, rfIgnoreCase);
  if _StartPos_ = 1 then
    Result := StringReplace(Text, _SubStr_, _Replacement_, rf)
  else
    Result := StringReplace(copy(Text, _StartPos_, MAXINT), _SubStr_, _Replacement_, rf);
end;

function DelphiPosFile;
var
  //ls: tstringlist;
  //error; using TstringList will change single $0D to $0D$0A pair
  fs: tfileStream;
  S: string;
begin
  //ls := TStringList.Create;
  fs := TFileStream.Create(_FileName_, default_fmRead);
  try
    try
      //ls.LoadFromFile(_FileName_);
      //Result := DelphiPos(ls.Text);
      SetLength(S, fs.Size);
      fs.position := 0;
      fs.read(S[1], fs.Size);
      Result := DelphiPos(S);
    except Result := 0; //raise;
    end
  finally fs.Free; //ls.Free;
  end;
end;


function DelphiCountFile;
var
  //ls: tstringlist;
  //error; using TstringList will change single $0D to $0D$0A pair
  fs: TFileStream;
  S: string;
begin
  //ls := TStringList.Create;
  fs := TFileStream.Create(_FileName_, default_fmRead);
  try
    try
      //ls.LoadFromFile(_FileName_);
      //Result := DelphiCount(ls.Text);
      SetLength(S, fs.Size);
      fs.position := 0;
      fs.read(S[1], fs.Size);
      Result := DelphiCount(S);
    except Result := 0; //raise;
    end
  finally fs.Free; //ls.Free;
  end;
end;

function DelphiReplaceFile;
var
  //ls: tstringlist;
  //error; using TstringList will change single $0D to $0D$0A pair
  fs: TFileStream;
  S: string;
begin
  //ls := TStringList.Create;
  fs := TFileStream.Create(_RFileName_, default_fmWrite);
  try
    try
      //ls.LoadFromFile(_RFileName_);
      //with ls do
      //  Text := DelphiReplace(Text);
      //ls.SaveToFile(_RFileName_);
      //Result := length(ls.Text);
      SetLength(S, fs.Size);
      fs.position := 0;
      fs.read(S[1], fs.Size);
      S := DelphiReplace(S);
      Result := length(S);
      fs.position := 0;
      fs.write(S[1], Result);
      fs.Size := Result;
    except Result := -1; //raise;
    end
  finally fs.free //ls.Free;
  end;
end;

function FastStringPos;
begin
  if _CaseSensitive_ then
    Result := FastStrings.FastPos(_S_, _SubStr_, length(_S_), length(_SubStr_), _StartPos_)
  else
    Result := FastStrings.FastPosNoCase(_S_, _SubStr_, length(_S_), length(_SubStr_), _StartPos_)
end;

function FastStringCount(const StartPos: integer): integer; overload; forward; //support for startpos

function FastStringCount(const Text: string): integer; overload; forward; //support for file access

function FastStringCount(const Text: string; StartPos: integer): integer; overload; forward; //support for file access

function FastStringCount: integer; overload;
// faststringcount originally does not have startpos feature
// we try not to alter its excellent performance too much
begin
  if _StartPos_ <= 1 then
    Result := FastStringFuncs.StringCount(_S_, _SubStr_, _CaseSensitive_)
  else
    Result := FastStringCount(_StartPos_);
end;

function FastStringCount(const StartPos: integer): integer; overload;
begin
  Result := FastStringFuncs.StringCount(copy(_S_, StartPos, MAXINT), _SubStr_, _CaseSensitive_)
end;

function FastStringCount(const Text: string): integer; overload;
begin
  if _StartPos_ <= 1 then
    Result := FastStringFuncs.StringCount(Text, _SubStr_, _CaseSensitive_)
  else
    Result := FastStringCount(Text, _StartPos_);
end;

function FastStringCount(const Text: string; StartPos: integer): integer; overload;
begin
  Result := FastStringFuncs.StringCount(copy(Text, StartPos, MAXINT), _SubStr_, _CaseSensitive_)
end;

function FastStringReplace(const StartPos: integer): string; overload; forward;

function FastStringReplace(const Text: string): string; overload; forward;

function FastStringReplace(const Text: string; const StartPos: integer): string; overload; forward;

function FastStringReplace: string; overload;
// faststringreplace originally does not have startpos feature
begin
  if _StartPos_ <= 1 then
    Result := FastStrings.FastReplace(_S_, _SubStr_, _Replacement_, _CaseSensitive_)
  else
    Result := FastStringReplace(_StartPos_)
end;

function FastStringReplace(const StartPos: integer): string; overload;
begin
  Result := Copy(_S_, 1, StartPos - 1) +
    FastStrings.FastReplace(copy(_S_, StartPos, MAXINT), _SubStr_, _Replacement_, _CaseSensitive_)
end;

function FastStringReplace(const Text: string): string; overload;
begin
  if _StartPos_ <= 1 then
    Result := FastStrings.FastReplace(Text, _SubStr_, _Replacement_, _CaseSensitive_)
  else
    Result := FastStringReplace(Text, _StartPos_);
end;

function FastStringReplace(const Text: string; const StartPos: integer): string; overload;
begin
  Result := Copy(Text, 1, StartPos - 1) +
    FastStrings.FastReplace(copy(Text, StartPos, MAXINT), _SubStr_, _Replacement_, _CaseSensitive_)
end;

function FastStringPosFile;
// faststrings originally does not have direct file access feature
var
  //ls: tstringlist;
  //error; using TstringList will change single $0D to $0D$0A pair
  fs: TFileStream;
  S: string;
begin
  //ls := TStringList.Create;
  fs := TFileStream.Create(_FileName_, default_fmRead);
  try
    try
      //ls.LoadFromFile(_FileName_);
      //with ls do
      //if _CaseSensitive_ then
      //  Result := FastStrings.FastPos(Text, _SubStr_, length(Text), length(_SubStr_), _StartPos_)
      //else
      //  Result := FastStrings.FastPosNoCase(Text, _SubStr_, length(Text), length(_SubStr_), _StartPos_)
      SetLength(S, fs.Size);
      fs.position := 0;
      fs.read(S[1], fs.Size);
      if _CaseSensitive_ then
        Result := FastStrings.FastPos(S, _SubStr_, length(S), length(_SubStr_), _StartPos_)
      else
        Result := FastStrings.FastPosNoCase(S, _SubStr_, length(S), length(_SubStr_), _StartPos_)
    except Result := 0; //raise;
    end
  finally fs.Free; //ls.Free;
  end;
end;

function FastStringCountFile;
// faststrings originally does not have direct file access feature
var
  //ls: tstringlist;
  //error; using TstringList will change single $0D to $0D$0A pair
  fs: TFileStream;
  S: string;
begin
  //ls := TStringList.Create;
  fs := TFileStream.Create(_FileName_, default_fmRead);
  try
    try
      //ls.LoadFromFile(_FileName_);
      //with ls do
      //  Result := FastStringCount(Text);
      SetLength(S, fs.Size);
      fs.position := 0;
      fs.read(S[1], fs.Size);
      Result := FastStringCount(S);
    except Result := 0; //raise;
    end
  finally fs.Free; //ls.Free;
  end;
end;

function FastStringReplaceFile;
var
  //ls: tstringlist;
  //error; using TstringList will change single $0D to $0D$0A pair
  fs: TFileStream;
  S: string;
begin
  //ls := TStringList.Create;
  fs := TFileStream.Create(_RFileName_, default_fmWrite);
  try
    try
      //ls.LoadFromFile(_RFileName_);
      //with ls do begin
      //  Text := FastStringReplace(Text);
      //  SaveToFile(_RFileName_);
      //  Result := length(Text);
      //end;
      SetLength(S, fs.Size);
      fs.position := 0;
      fs.read(S[1], fs.Size);
      S := FastStringReplace(S);
      Result := length(S);
      fs.position := 0;
      fs.write(S[1], Result);
      fs.Size := Result;
    except Result := -1; //raise;
    end
  finally fs.Free; //ls.Free;
  end;
end;

function AJSearchPos;
// note: AJSearch FindFrom Result (maybe) is 0-based
var
  AJStartPos: integer;
begin
  with TSearch.Create(Self) do begin
    try
      Pattern := _SubStr_;
      CaseSensitive := _CaseSensitive_;
      SetData(pChar(_S_), length(_S_));
      //Result := FindFrom(_StartPos_);
      AJStartPos := _StartPos_ - 1;
      Result := FindFrom(AJStartPos);
    finally free
    end;
  end;
end;

function AJSearchCount;
// note: AJSearch Count Result (maybe) is 0-based
begin
  with TSearch.Create(Self) do begin
    try
      Pattern := _SubStr_;
      CaseSensitive := _CaseSensitive_;
      SetData(pChar(_S_), length(_S_));
      if FindFrom(_StartPos_ - 1) >= 0 //if FindFrom(_StartPos_)
        then Result := 1
      else Result := 0;
      if Result >= 0 then //if Result > 0 then
        while FindNext >= 0 do inc(Result) //while FindNext > 0 do inc(Result)
    finally free
    end;
  end;
end;

function AJSearchReplace: string; overload;
// AJSearch originally does not have Replace feature
var
  fPos, LastPos, SubLen: integer;
begin
  with TSearch.Create(Self) do begin
    try
      Pattern := _SubStr_;
      CaseSensitive := _CaseSensitive_;
      SetData(pChar(_S_), length(_S_));
      fPos := FindFrom(_StartPos_ - 1); //fPos := FindFrom(_StartPos_);
      if fPos < 0 then //if fPos < 1 then
        Result := _S_
      else begin
        Result := ''; // failing to do this will result in a bit headache
        LastPos := 0; //LastPos := 1;
        SubLen := length(_SubStr_);
        repeat
          //Result := Result + copy(_S_, LastPos, fPos - LastPos) + _Replacement_;
          Result := Result + copy(_S_, (LastPos + 1), fPos - LastPos) + _Replacement_;
          LastPos := fPos + SubLen;
          fPos := FindNext;
        until fPos < 0; //until fPos < 1;
        //Result := Result + copy(_S_, (LastPos+1), MAXINT);
        Result := Result + copy(_S_, (LastPos + 1), MAXINT);
      end;
    finally free
    end;
  end;
end;

function AJSearchReplace(const _S_: string): string; overload;
// AJSearch originally does not have Replace feature
// note: this routine Used by AJSearch Replace in File
var
  fPos, LastPos, SubLen: integer;
begin
  with TSearch.Create(Self) do begin
    try
      Pattern := _SubStr_;
      CaseSensitive := _CaseSensitive_;
      SetData(pChar(_S_), length(_S_));
      fPos := FindFrom(_StartPos_ - 1); //fPos := FindFrom(_StartPos_);
      if fPos < 0 then //if fPos < 1 then
        Result := _S_
      else begin
        Result := ''; // failing to do this will result in a bit headache
        LastPos := 0; //LastPos := 1;
        SubLen := length(_SubStr_);
        repeat
          //Result := Result + copy(_S_, LastPos, fPos - LastPos) + _Replacement_;
          Result := Result + copy(_S_, (LastPos + 1), fPos - LastPos) + _Replacement_;
          LastPos := fPos + SubLen;
          fPos := FindNext;
        until fPos < 0; //until fPos < 1;
        //Result := Result + copy(_S_, (LastPos+1), MAXINT);
        Result := Result + copy(_S_, (LastPos + 1), MAXINT);
      end;
    finally free
    end;
  end;
end;
function AJSearchPosFile;
// note: AJSearch FindFrom Result (maybe) is 0-based
begin
  with TFileSearch.Create(Self) do begin
    try
      FileName := _FileName_;
      CaseSensitive := _CaseSensitive_;
      Pattern := _SubStr_;
      Result := FindFrom(_StartPos_ - 1)
    finally free
    end;
  end;
end;

function AJSearchCountFile;
// AJSearch originally does not have routine count in file
begin
  with TFileSearch.Create(Self) do begin
    try
      FileName := _FileName_;
      Pattern := _SubStr_;
      CaseSensitive := _CaseSensitive_;
      //if FindFrom(_StartPos_) > 0 then
      if FindFrom(_StartPos_) >= 0 then
        Result := 1
      else Result := 0;
      if Result > 0 then
        //while FindNext > 0 do
        while FindNext >= 0 do
          inc(Result)
    finally free;
    end;
  end;
end;

function AJSearchReplaceFile;
// AJSearch originally does not have direct file access replace
var
  ls: TStringList;
begin
  if AJSearchCountFile < 1 then
    with TFileStream.Create(_RFileName_, fmOpenRead or fmShareDenyNone) do begin
      Result := Size;
      Free;
    end
  else begin
    ls := TStringList.Create;
    try
      try
        ls.LoadFromFile(_RFilename_);
        with ls do
          Text := AJSearchReplace(Text);
        ls.SaveToFile(_RFilename_);
        Result := length(ls.Text);
      except Result := -1;
      end;
    finally
      ls.Free;
    end;
  end;
  //with TFileSearch.Create(Self) do begin
  //  try
  //    FileName := _RFileName_;
  //    Pattern := _SubStr_;
  //    CaseSensitive := _CaseSensitive_;
  //    Result := FindFrom(_StartPos_)
  //  finally free
  //  end;
  //end;
end;

function MWSearchPos(const mw: TmTSearch; const S: string): integer; overload;
// note: MWSearch Result is not reliable
begin
  with mw do begin
    try
      try
        if _CaseSensitive_ then begin
          Init(_SubStr_);
          Result := FindFirst(S);
          while (Result > 0) and (Result < _StartPos_) do
            Result := Next;
        end
        else begin
          iInit(_SubStr_);
          Result := iFindFirst(S);
          while (Result > 0) and (Result < _StartPos_) do
            Result := iNext;
        end;
      except Result := 0
      end;
    finally //free ..oops :)
    end;
  end;
end;

function MWSearchPos(const S: string): integer; overload;
// note: MWSearch Result is not reliable
begin
  with TmTSearch.Create do begin
    try
      try
        if _CaseSensitive_ then begin
          Init(_SubStr_);
          Result := FindFirst(S);
          while (Result > 0) and (Result < _StartPos_) do
            Result := Next;
        end
        else begin
          iInit(_SubStr_);
          Result := iFindFirst(S);
          while (Result > 0) and (Result < _StartPos_) do
            Result := iNext;
        end;
      except Result := 0
      end;
    finally free
    end;
  end;
end;

function MWSearchPos: integer; overload;
begin
  Result := MWSearchPos(_S_);
end;

function MWSearchCount(const S: string): integer; overload;
// note: MWSearch Result is not reliable
var
  mw: TmTSearch;
begin
  mw := TmTSearch.Create;
  try
    try
      Result := 0;
      if MWSearchPos(mw, S) > 0 then
        if _CaseSensitive_ then
          repeat inc(Result); until mw.Next = 0
        else
          repeat inc(Result); until mw.iNext = 0
    except Result := 0
    end;
  finally mw.free
  end;
end;

function MWSearchCount: integer; overload;
// note: MWSearch Result is not reliable
begin
  Result := MWSearchCount(_S_);
end;

function MWSearchReplace(const S: string): string; overload;
// note: MWSearch Result is not reliable
var
  fPos, LastPos, SubLen: integer;
  mwInit: procedure(S: string) of object;
  mwFindFirst: function(S: string): integer of object;
  mwNext: function: integer of object;
begin
  with TmTSearch.Create do begin
    try
      try
        if _CaseSensitive_ then begin
          mwInit := Init;
          mwFIndFirst := FindFirst;
          mwNext := Next;
        end else begin
          mwInit := iInit;
          mwFIndFirst := iFindFirst;
          mwNext := iNext;
        end;
        mwInit(_SubStr_);
        fPos := mwFindFirst(S);
        while (fPos > 0) and (fPos < _StartPos_) do fPos := mwNext;
        if fPos < 1 then Result := S
        else begin
          Result := ''; // do not fail to do this
          LastPos := 1;
          SubLen := length(_SubStr_);
          repeat
            Result := Result + copy(S, LastPos, fPos - LastPos) + _Replacement_;
            LastPos := fPos + SubLen;
            fPos := mwNext;
          until fPos < 1;
          Result := Result + copy(S, LastPos, MAXINT);
        end;
      except result := ''
      end;
    finally free
    end;
  end;
end;

function MWSearchReplace: string; overload;
begin
  Result := MWSearchReplace(_S_);
end;

function MWSearchPosFile;
// note: MWSearch Result is not reliable
var
  ls: tstringlist;
begin
  ls := TStringList.Create;
  try
    ls.LoadFromFile(_FileName_);
    Result := MWSearchPos(ls.Text);
  finally ls.free
  end;
end;

function MWSearchCountFile;
// note: MWSearch Result is not reliable
var
  ls: tstringlist;
begin
  ls := TStringList.Create;
  try
    ls.LoadFromFile(_FileName_);
    Result := MWSearchCount(ls.Text);
  finally ls.free
  end;
end;

function MWSearchReplaceFile;
// note: MWSearch Result is not reliable
var
  mw: TmTSearch;
  ls: tstringlist;
begin
  //Result := -1; exit;
  mw := TmTSearch.Create;
  ls := TStringList.Create;
  try
    try
      ls.LoadFromFile(_RFileName_);
      with ls do
        Text := MWSearchReplace(Text);
      ls.SaveToFile(_RFileName_);
      Result := length(ls.Text);
    except Result := -1;
    end;
  finally ls.free; mw.free
  end;
end;

function cxposPos;
begin
  with cxpos.txSearch.Create(_SubStr_, _IgnoreCase_) do begin
    try
      Result := Pos(_S_, _StartPos_);
    finally free
    end;
  end;
end;

function cxposCount;
begin
  with cxpos.txSearch.Create(_SubStr_, _IgnoreCase_) do begin
    try
      Result := Count(_S_, _StartPos_);
    finally free
    end;
  end;
end;

function cxposReplace;
begin
  with cxpos.txSearch.Create(_SubStr_, _IgnoreCase_) do begin
    try
      Result := StrReplace(_S_, _Replacement_, _StartPos_);
    finally free
    end;
  end;
end;

function cxposPosFile;
begin
  with cxpos.txSearch.Create(_SubStr_, _IgnoreCase_) do begin
    try
      Result := FindFirst(_FileName_, _StartPos_);
    finally free
    end;
  end;
end;

function cxposCountFile;
begin
  with cxpos.txSearch.Create(_SubStr_, _IgnoreCase_) do begin
    try
      Result := InFileCount(_FileName_, _StartPos_);
    finally free
    end;
  end;
end;

function cxposReplaceFile;
begin
  with cxpos.txSearch.Create(_SubStr_, _IgnoreCase_) do begin
    try
      Result := InFileReplace(_RFileName_, _Replacement_, _StartPos_);
    finally free
    end;
  end;
end;

procedure InitGlobalSettings;
begin
  with GlobalSettings do begin
    EnabledContestants := DefaultEnabledContestants;
    CheckedMethodes := DefaultCheckedMethodes;
  end;
end;

procedure initContestantfunctions;
//TContestant = (ctDelphi, ctFastString, ctAJSearch, ctMWSearch, ctcxpos);
begin
  with ContestantFunctions[ctDelphi] do begin
    fnPos := DelphiPos;
    fnCount := DelphiCount;
    fnReplace := DelphiReplace;
    fnPosFile := DelphiPosFile;
    fnCountFile := DelphiCountFile;
    fnReplaceFile := DelphiReplaceFile;
  end;
  with ContestantFunctions[ctFastString] do begin
    fnPos := FastStringPos;
    fnCount := FastStringCount;
    fnReplace := FastStringReplace;
    fnPosFile := FastStringPosFile;
    fnCountFile := FastStringCountFile;
    fnReplaceFile := FastStringReplaceFile;
  end;
  with ContestantFunctions[ctAJSearch] do begin
    fnPos := AJSearchPos;
    fnCount := AJSearchCount;
    fnReplace := AJSearchReplace;
    fnPosFile := AJSearchPosFile;
    fnCountFile := AJSearchCountFile;
    fnReplaceFile := AJSearchReplaceFile;
  end;
  with ContestantFunctions[ctMWSearch] do begin
    fnPos := MWSearchPos;
    fnCount := MWSearchCount;
    fnReplace := MWSearchReplace;
    fnPosFile := MWSearchPosFile;
    fnCountFile := MWSearchCountFile;
    fnReplaceFile := MWSearchReplaceFile;
  end;
  with ContestantFunctions[ctcxpos] do begin
    fnPos := cxposPos;
    fnCount := cxposCount;
    fnReplace := cxposReplace;
    fnPosFile := cxposPosFile;
    fnCountFile := cxposCountFile;
    fnReplaceFile := cxposReplaceFile;
  end;
end;

initialization
  initContestantfunctions;
  InitGlobalSettings;
end.

